from __future__ import annotations

import contextlib
import itertools
from collections import defaultdict
from collections.abc import Generator, Iterable, Mapping, Sequence
from copy import deepcopy
from dataclasses import dataclass
from functools import lru_cache
from typing import Final, Optional

from rflx import expression as expr, ir, typing_ as rty
from rflx.common import Base, indent, indent_next, verbose_repr
from rflx.error import Location, RecordFluxError, Severity, Subsystem
from rflx.identifier import ID, StrID, id_generator

from . import (
    declaration as decl,
    statement as stmt,
    type_ as mty,
)
from .message import Message, Refinement
from .top_level_declaration import TopLevelDeclaration, UncheckedTopLevelDeclaration


class Transition(Base):
    def __init__(
        self,
        target: StrID,
        condition: expr.Expr = expr.TRUE,
        description: Optional[str] = None,
        location: Optional[Location] = None,
    ):
        self.target = ID(target)
        self.condition = condition
        self.description = description
        self.location = location

    def __repr__(self) -> str:
        return verbose_repr(self, ["target", "condition", "description"])

    def __str__(self) -> str:
        target = self.target if self.target != FINAL_STATE.identifier else "null"
        with_aspects = f'\n   with Desc => "{self.description}"' if self.description else ""
        if_condition = (
            f"\n   if {indent_next(str(self.condition), 6)}" if self.condition != expr.TRUE else ""
        )
        return f"goto {target}{with_aspects}{if_condition}"

    def to_ir(self, variable_id: Generator[ID, None, None]) -> ir.Transition:
        condition = self.condition.to_ir(variable_id)
        return ir.Transition(
            self.target,
            ir.ComplexExpr(
                condition.stmts,
                condition.expr,
            ),
            self.description,
            self.location,
        )


class State(Base):
    def __init__(  # noqa: PLR0913
        self,
        identifier: StrID,
        transitions: Optional[Sequence[Transition]] = None,
        exception_transition: Optional[Transition] = None,
        actions: Optional[Sequence[stmt.Statement]] = None,
        declarations: Optional[Sequence[decl.BasicDeclaration]] = None,
        description: Optional[str] = None,
        location: Optional[Location] = None,
    ):
        if transitions:
            assert transitions[-1].condition == expr.TRUE, "missing default transition"
        else:
            assert not actions
            assert not declarations

        assert exception_transition.condition == expr.TRUE if exception_transition else True

        self._identifier = ID(identifier)
        self._transitions = transitions or []
        self._exception_transition = exception_transition
        self._actions = actions or []
        self.declarations = {d.identifier: d for d in declarations} if declarations else {}
        self.description = description
        self.location = location

        self._normalize()

    def __repr__(self) -> str:
        return verbose_repr(self, ["identifier", "transitions", "actions", "declarations"])

    def __str__(self) -> str:
        with_aspects = f'\n   with Desc => "{self.description}"\n' if self.description else " "
        if not self.declarations and not self.actions and not self.transitions:
            return f"state {self.identifier}{with_aspects}is null state"
        declarations = "".join([f"{d};\n" for d in self.declarations.values()])
        actions = "".join([f"{a};\n" for a in self.actions])
        transitions = "\n".join([f"{p}" for p in self.transitions])
        exception_transition = (
            f"\nexception\n   {self.exception_transition}" if self.exception_transition else ""
        )
        return (
            f"state {self.identifier}{with_aspects}is\n{indent(declarations, 3)}begin\n"
            f"{indent(actions, 3)}"
            f"transition\n{indent(transitions, 3)}{indent_next(exception_transition, 0)}\n"
            f"end {self.identifier}"
        )

    @property
    def identifier(self) -> ID:
        return self._identifier

    @property
    def transitions(self) -> Sequence[Transition]:
        return self._transitions or []

    @property
    def exception_transition(self) -> Optional[Transition]:
        return self._exception_transition

    @property
    def actions(self) -> Sequence[stmt.Statement]:
        return self._actions

    @property
    def has_exceptions(self) -> bool:
        # TODO(eng/recordflux/RecordFlux#861): Move into IR
        # The need for exception transitions should be made dependent on the presence of asserts in
        # the IR.
        def has_expression_exceptions(expression: expr.Expr) -> bool:
            return bool(
                expression.findall(
                    lambda x: isinstance(
                        x,
                        (
                            expr.Selected,
                            expr.Head,
                            expr.Comprehension,
                            expr.MessageAggregate,
                            expr.DeltaMessageAggregate,
                            expr.Conversion,
                            expr.Opaque,
                            expr.Add,
                            expr.Sub,
                            expr.Mul,
                            expr.Div,
                            expr.Mod,
                            expr.Pow,
                        ),
                    ),
                ),
            )

        return any(
            isinstance(a, (stmt.Append, stmt.Extend, stmt.MessageFieldAssignment))
            or (
                isinstance(a, stmt.VariableAssignment)
                and (isinstance(a.type_, rty.Message) or (has_expression_exceptions(a.expression)))
            )
            for a in self._actions
        ) or any(has_expression_exceptions(t.condition) for t in self._transitions)

    def optimize(self) -> None:
        self._optimize_structures()

    def to_ir(self, variable_id: Generator[ID, None, None]) -> ir.State:
        actions = [s for a in self._actions for s in a.to_ir(variable_id)]
        transitions = [t.to_ir(variable_id) for t in self._transitions]
        declarations = [d.to_ir(variable_id) for d in self.declarations.values()]

        return ir.State(
            self.identifier,
            transitions,
            (self.exception_transition.to_ir(variable_id) if self.exception_transition else None),
            [*declarations, *actions],
            self.description,
            self.location,
        )

    def _normalize(self) -> None:
        self._normalize_transitions()
        self._normalize_actions()

    def _normalize_transitions(self) -> None:
        for t in self._transitions:
            if t.target == ID("null"):
                t.target = FINAL_STATE.identifier
        if self._exception_transition and self._exception_transition.target == ID("null"):
            self._exception_transition.target = FINAL_STATE.identifier

    def _normalize_actions(self) -> None:  # noqa: PLR0912
        field_assignments: list[stmt.MessageFieldAssignment] = []
        actions: list[stmt.Statement] = []

        for a in self._actions:
            if not field_assignments:
                if isinstance(a, stmt.MessageFieldAssignment):
                    field_assignments.append(a)
                else:
                    actions.append(a)
            elif len(field_assignments) == 1:
                if isinstance(a, stmt.MessageFieldAssignment):
                    if field_assignments[0].message == a.message:
                        field_assignments.append(a)
                    else:
                        actions.extend(field_assignments)
                        field_assignments = [a]
                else:
                    actions.extend(
                        [
                            *field_assignments,
                            a,
                        ],
                    )
                    field_assignments = []
            else:
                if isinstance(a, stmt.MessageFieldAssignment):
                    if field_assignments[0].message == a.message:
                        field_assignments.append(a)
                    else:
                        actions.append(self._collect_message_field_assignments(field_assignments))
                        field_assignments = [a]
                else:
                    actions.extend(
                        [
                            self._collect_message_field_assignments(field_assignments),
                            a,
                        ],
                    )
                    field_assignments = []

        if len(field_assignments) == 1:
            actions.extend(field_assignments)
        elif len(field_assignments) > 1:
            actions.append(self._collect_message_field_assignments(field_assignments))

        self._actions = actions

    def _collect_message_field_assignments(
        self,
        field_assignments: list[stmt.MessageFieldAssignment],
    ) -> stmt.VariableAssignment:
        return stmt.VariableAssignment(
            field_assignments[0].identifier,
            expr.DeltaMessageAggregate(
                field_assignments[0].identifier,
                {a.field: a.expression for a in field_assignments},
            ),
            location=Location(
                field_assignments[0].location.start if field_assignments[0].location else (0, 0),
                field_assignments[0].location.source if field_assignments[0].location else None,
                field_assignments[-1].location.end if field_assignments[-1].location else None,
            ),
        )

    def _optimize_structures(self) -> None:
        """
        Replace local messages with structures if possible.

        If only a limited feature set of a local message object is used it can be replaced with
        a structure. This allows the generation of more size and runtime efficient code.
        """

        def find_identifier(name: ID, expression: expr.Expr) -> bool:
            return bool(
                expression.findall(lambda e: isinstance(e, expr.Variable) and e.identifier == name),
            )

        def find_attribute_prefix(name: ID, expression: expr.Expr) -> bool:
            return bool(
                expression.findall(
                    lambda e: isinstance(e, expr.Attribute) and find_identifier(name, e.prefix),
                ),
            )

        def substituted(expression: expr.Expr, structure: rty.Structure) -> expr.Expr:
            def replace_expression_type(expression: expr.Expr) -> expr.Expr:
                if (
                    isinstance(expression, (expr.Variable, expr.Call))
                    and isinstance(expression.type_, rty.Message)
                    and expression.type_.identifier == structure.identifier
                ):
                    expression.type_ = structure
                return expression

            return expression.substituted(replace_expression_type)

        def contains_unsupported_feature(name: ID, action: stmt.Statement) -> bool:
            return (
                (isinstance(action, stmt.Append) and find_identifier(name, action.parameter))
                or (isinstance(action, stmt.AttributeStatement) and action.identifier == name)
                or (isinstance(action, stmt.MessageFieldAssignment) and action.message == name)
                or (
                    isinstance(action, stmt.VariableAssignment)
                    and action.identifier == name
                    and not isinstance(action.expression, expr.Call)
                )
                or (
                    isinstance(action, stmt.Assignment)
                    and find_attribute_prefix(name, action.expression)
                )
            )

        for name, declaration in self.declarations.items():
            if (
                not isinstance(declaration, decl.VariableDeclaration)
                or not isinstance(declaration.type_, rty.Message)
                or not declaration.type_.is_definite
                or (
                    isinstance(declaration, decl.VariableDeclaration)
                    and declaration.expression is not None
                )
            ):
                continue

            for action in self._actions:
                if contains_unsupported_feature(name, action):
                    break
            else:
                message_decl = decl.VariableDeclaration(
                    identifier=declaration.identifier,
                    type_identifier=declaration.type_identifier,
                    expression=declaration.expression,
                    type_=rty.Structure(
                        identifier=declaration.type_.identifier,
                        field_combinations=declaration.type_.field_combinations,
                        parameter_types=declaration.type_.parameter_types,
                        field_types=declaration.type_.field_types,
                    ),
                    location=declaration.location,
                )
                assert isinstance(message_decl.type_, rty.Structure)
                self.declarations[name] = message_decl

                for action in self._actions:
                    if isinstance(action, stmt.Assignment):
                        substituted(action.expression, message_decl.type_)

                for transition in self._transitions:
                    substituted(transition.condition, message_decl.type_)


class Session(TopLevelDeclaration):
    def __init__(  # noqa: PLR0913
        self,
        identifier: StrID,
        states: Sequence[State],
        declarations: Sequence[decl.BasicDeclaration],
        parameters: Sequence[decl.FormalDeclaration],
        types: Sequence[mty.Type],
        location: Optional[Location] = None,
        workers: int = 1,
    ):
        super().__init__(identifier, location)

        self.error.propagate()

        self.states = [*states, FINAL_STATE] if FINAL_STATE not in states else states
        self.declarations = {d.identifier: d for d in declarations}
        self.parameters = {p.identifier: p for p in parameters}
        self.direct_dependencies = {t.identifier: t for t in types}
        self.types = self.direct_dependencies.copy()
        self.location = location
        self._workers = workers

        refinements = [t for t in types if isinstance(t, Refinement)]

        for t in types:
            if isinstance(t, Message):
                t.set_refinements([r for r in refinements if r.pdu == t])

        assert all(not isinstance(d, decl.FormalDeclaration) for d in self.declarations.values())

        self._global_declarations: Mapping[ID, decl.Declaration] = {
            **self.parameters,
            **self.declarations,
        }
        self._enum_literals = mty.enum_literals(self.types.values(), self.package)
        self._type_names = mty.qualified_type_names(self.types.values())

        self._check_identifiers()
        self._normalize()
        self._validate()

        self.error.propagate()

        self._optimize()
        self.to_ir()

    def __hash__(self) -> int:
        return hash(self.identifier)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.identifier == other.identifier
                and self.states == other.states
                and self.declarations == other.declarations
                and self.parameters == other.parameters
                and self.types == other.types
            )
        return NotImplemented

    def __repr__(self) -> str:
        return verbose_repr(self, ["identifier", "states", "declarations", "parameters"])

    def __str__(self) -> str:
        parameters = "".join([f"{p};\n" for p in self.parameters.values()])
        declarations = "".join([f"{d};\n" for d in self.declarations.values()])
        states = "\n\n".join([f"{s};" for s in self.states if s != FINAL_STATE])
        return (
            f"generic\n{indent(parameters, 3)}session {self.identifier.name} is\n"
            f"{indent(declarations, 3)}begin\n{indent(states, 3)}\nend {self.identifier.name}"
        )

    @property
    def initial_state(self) -> State:
        return self.states[0]

    @lru_cache  # noqa: B019
    def to_ir(self) -> ir.Session:
        variable_id = id_generator()
        return ir.Session(
            self.identifier,
            [state.to_ir(variable_id) for state in self.states],
            [d.to_ir(variable_id) for d in self.declarations.values()],
            [p.to_ir() for p in self.parameters.values()],
            self.types,
            self.location,
            variable_id,
            self._workers,
        )

    def _normalize(self) -> None:  # noqa: PLR0912
        """
        Normalize all expressions of the session.

        - Unify the identifier casing.
        - Replace variables by type names, literals or function calls without arguments where
          necessary. The distinction between these different kind of expressions is not possible in
          the parser, as all of these are syntactically identical.
        """
        functions = [
            p.identifier
            for p in self.parameters.values()
            if isinstance(p, decl.FunctionDeclaration)
        ]
        channels = [
            p.identifier for p in self.parameters.values() if isinstance(p, decl.ChannelDeclaration)
        ]
        global_variables = list(self.declarations)
        type_names_map = {t: t for t in self._type_names}

        def normalize_identifiers_global(expression: expr.Expr) -> expr.Expr:
            return normalize_identifiers(
                expression,
                global_variables,
                self._enum_literals,
                self._type_names,
                functions,
            )

        for d in self.declarations.values():
            if isinstance(d, decl.TypeCheckableDeclaration) and d.type_identifier in type_names_map:
                d.type_identifier = ID(
                    type_names_map[d.type_identifier],
                    location=d.type_identifier.location,
                )

            if isinstance(d, decl.VariableDeclaration) and d.expression:
                d.expression = d.expression.substituted(normalize_identifiers_global)

        states_map = {s.identifier: s.identifier for s in self.states}

        for state in self.states:
            variables = [*global_variables, *state.declarations]
            declarations_map = {
                **{v: v for v in variables},
                **{c: c for c in channels},
            }

            def normalize_identifiers_local(
                expression: expr.Expr,
                variables: Iterable[ID] = variables,
            ) -> expr.Expr:
                return normalize_identifiers(
                    expression,
                    variables,
                    self._enum_literals,
                    self._type_names,
                    functions,
                )

            for d in state.declarations.values():
                if isinstance(d, decl.VariableDeclaration) and d.expression:
                    d.expression = d.expression.substituted(normalize_identifiers_local)
            for a in state.actions:
                if isinstance(a, stmt.Assignment):
                    a.expression = a.expression.substituted(normalize_identifiers_local)
                if isinstance(a, stmt.AttributeStatement):
                    if a.identifier in declarations_map:
                        a.identifier = ID(
                            declarations_map[a.identifier],
                            location=a.identifier.location,
                        )
                    a.parameters = [
                        p.substituted(normalize_identifiers_local) for p in a.parameters
                    ]
                if isinstance(a, stmt.Reset):
                    a.associations = {
                        i: e.substituted(normalize_identifiers_local)
                        for i, e in a.associations.items()
                    }
            for t in state.transitions:
                if t.target in states_map:
                    t.target = states_map[t.target]
                t.condition = t.condition.substituted(normalize_identifiers_local)

            if state.exception_transition and t.target in states_map:
                state.exception_transition.target = states_map[state.exception_transition.target]

    def _optimize(self) -> None:
        for state in self.states:
            state.optimize()

    def _check_identifiers(self) -> None:
        self.error.extend(
            mty.check_identifier_notation(
                itertools.chain(
                    (
                        expr.Variable(d.type_identifier)
                        for d in self.declarations.values()
                        if isinstance(d, decl.TypeCheckableDeclaration)
                    ),
                    (
                        d.expression
                        for d in self.declarations.values()
                        if isinstance(d, decl.VariableDeclaration) and d.expression
                    ),
                ),
                itertools.chain(
                    self.parameters,
                    self.declarations,
                    self._enum_literals,
                    self._type_names,
                ),
            ),
        )

        for state in self.states:
            self.error.extend(
                mty.check_identifier_notation(
                    itertools.chain(
                        (
                            d.expression
                            for d in state.declarations.values()
                            if isinstance(d, decl.VariableDeclaration) and d.expression
                        ),
                        (a.expression for a in state.actions if isinstance(a, stmt.Assignment)),
                        (
                            e
                            for a in state.actions
                            if isinstance(a, stmt.AttributeStatement)
                            for e in [expr.Variable(a.identifier), *a.parameters]
                        ),
                        (
                            e
                            for a in state.actions
                            if isinstance(a, stmt.Reset)
                            for e in a.associations.values()
                        ),
                        (
                            e
                            for t in state.transitions
                            for e in [expr.Variable(t.target), t.condition]
                        ),
                        (
                            expr.Variable(t.target)
                            for t in [state.exception_transition]
                            if t is not None
                        ),
                    ),
                    itertools.chain(
                        self.parameters,
                        self.declarations,
                        self._enum_literals,
                        self._type_names,
                        state.declarations,
                        (s.identifier for s in self.states),
                    ),
                ),
            )

    def _validate_states(self) -> None:
        if all(s == FINAL_STATE for s in self.states):
            self.error.extend(
                [("empty states", Subsystem.MODEL, Severity.ERROR, self.location)],
            )

        self._validate_state_existence()
        self._validate_duplicate_states()
        self._validate_state_reachability()

    def _validate_state_existence(self) -> None:
        state_identifiers = {s.identifier for s in self.states}
        for s in self.states:
            for t in s.transitions:
                if t.target not in state_identifiers:
                    self.error.extend(
                        [
                            (
                                f'transition from state "{s.identifier}" to non-existent state'
                                f' "{t.target}" in "{self.identifier}"',
                                Subsystem.MODEL,
                                Severity.ERROR,
                                t.target.location,
                            ),
                        ],
                    )

    def _validate_duplicate_states(self) -> None:
        identifier_states = defaultdict(list)
        for s in self.states:
            identifier_states[s.identifier].append(s)

        for identifier, states in identifier_states.items():
            if len(states) >= 2:
                for s in states[1:]:
                    self.error.extend(
                        [
                            (
                                f'duplicate state "{identifier}"',
                                Subsystem.MODEL,
                                Severity.ERROR,
                                s.location,
                            ),
                            (
                                f'previous definition of state "{identifier}"',
                                Subsystem.MODEL,
                                Severity.INFO,
                                states[0].location,
                            ),
                        ],
                    )

    def _validate_state_reachability(self) -> None:
        inputs: dict[ID, list[ID]] = {}
        for s in self.states:
            for t in [
                *s.transitions,
                *([s.exception_transition] if s.exception_transition else []),
            ]:
                if t.target in inputs:
                    inputs[t.target].append(s.identifier)
                else:
                    inputs[t.target] = [s.identifier]

            if s not in [self.initial_state, FINAL_STATE] and s.identifier not in inputs:
                self.error.extend(
                    [
                        (
                            f'unreachable state "{s.identifier}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            s.location,
                        ),
                    ],
                )

            if s != FINAL_STATE and not s.transitions:
                self.error.extend(
                    [
                        (
                            f'detached state "{s.identifier}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            s.location,
                        ),
                    ],
                )

    def _validate_declarations(
        self,
        declarations: Mapping[ID, decl.Declaration],
        visible_declarations: Mapping[ID, decl.Declaration],
    ) -> None:
        visible_declarations = dict(visible_declarations)

        def undefined_type(type_identifier: StrID, location: Optional[Location]) -> None:
            self.error.extend(
                [
                    (
                        f'undefined type "{type_identifier}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        location,
                    ),
                ],
            )

        for k, d in declarations.items():
            if k in visible_declarations:
                self.error.extend(
                    [
                        (
                            f'local variable "{k}" shadows previous declaration',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            d.location,
                        ),
                        (
                            f'previous declaration of variable "{k}"',
                            Subsystem.MODEL,
                            Severity.INFO,
                            visible_declarations[k].location,
                        ),
                    ],
                )

            self._reference_variable_declaration(d.variables(), visible_declarations)

            if isinstance(d, decl.TypeCheckableDeclaration):
                type_identifier = mty.internal_type_identifier(d.type_identifier, self.package)
                if type_identifier in self.types:
                    self.error.extend(
                        d.check_type(
                            self.types[type_identifier].type_,
                            lambda x: self._typify_variable(x, visible_declarations),
                        ),
                    )
                else:
                    undefined_type(d.type_identifier, d.location)
                    d.type_ = rty.Any()

                if isinstance(d, decl.FunctionDeclaration):
                    for a in d.arguments:
                        type_identifier = mty.internal_type_identifier(
                            a.type_identifier,
                            self.package,
                        )
                        if type_identifier in self.types:
                            a.type_ = self.types[type_identifier].type_
                        else:
                            a.type_ = rty.Any()
                            undefined_type(a.type_identifier, d.location)

            visible_declarations[k] = d

    def _validate_actions(
        self,
        actions: Sequence[stmt.Statement],
        declarations: Mapping[ID, decl.Declaration],
        local_declarations: Mapping[ID, decl.Declaration],
    ) -> None:
        self._validate_io_states(actions, local_declarations)

        for a in actions:
            try:
                type_ = declarations[a.identifier].type_
            except KeyError:
                type_ = rty.Undefined()

            self.error.extend(
                a.check_type(
                    type_,
                    lambda x: self._typify_variable(x, declarations),
                ),
            )

            self._reference_variable_declaration(a.variables(), declarations)

            if isinstance(a, stmt.Assignment):
                a.expression.substituted(lambda e: error_on_unsupported_expression(e, self.error))
            else:
                assert isinstance(a, stmt.AttributeStatement)
                for e in a.parameters:
                    e.substituted(lambda e: error_on_unsupported_expression(e, self.error))

    def _validate_io_states(
        self,
        actions: Sequence[stmt.Statement],
        local_declarations: Mapping[ID, decl.Declaration],
    ) -> None:
        io_statements = [a for a in actions if isinstance(a, stmt.ChannelAttributeStatement)]

        if io_statements:
            if local_declarations:
                self.error.extend(
                    [
                        (
                            "IO state must not contain declarations",
                            Subsystem.MODEL,
                            Severity.ERROR,
                            next(iter(local_declarations.values())).location,
                        ),
                    ],
                )

            if len(io_statements) != len(actions):
                self.error.extend(
                    [
                        (
                            "channel IO must not be combined with other actions in one state",
                            Subsystem.MODEL,
                            Severity.ERROR,
                            io_statements[0].location,
                        ),
                    ],
                )

        for i, s1 in enumerate(io_statements):
            if not isinstance(s1.parameter, expr.Variable):
                self.error.extend(
                    [
                        (
                            "channel parameter must be a variable",
                            Subsystem.MODEL,
                            Severity.ERROR,
                            s1.location,
                        ),
                    ],
                )

            for j, s2 in enumerate(io_statements):
                if i >= j:
                    continue

                if s1.identifier == s2.identifier:
                    self.error.extend(
                        [
                            (
                                f'channel "{s1.identifier}" may be read or written'
                                " at most once per state",
                                Subsystem.MODEL,
                                Severity.ERROR,
                                s1.identifier.location,
                            ),
                        ],
                    )
                    self.error.extend(
                        [
                            (
                                "conflicting read/write",
                                Subsystem.MODEL,
                                Severity.INFO,
                                s2.identifier.location,
                            ),
                        ],
                    )

                if (
                    isinstance(s1.parameter, expr.Variable)
                    and isinstance(s2.parameter, expr.Variable)
                    and s1.parameter.identifier == s2.parameter.identifier
                ):
                    self.error.extend(
                        [
                            (
                                f'message "{s1.parameter.identifier}" may be read or written'
                                " at most once per state",
                                Subsystem.MODEL,
                                Severity.ERROR,
                                s1.parameter.location,
                            ),
                        ],
                    )
                    self.error.extend(
                        [
                            (
                                "conflicting read/write",
                                Subsystem.MODEL,
                                Severity.INFO,
                                s2.parameter.location,
                            ),
                        ],
                    )

    def _validate_transitions(
        self,
        state: State,
        declarations: Mapping[ID, decl.Declaration],
    ) -> None:
        for t in state.transitions:
            t.condition = t.condition.substituted(lambda x: self._typify_variable(x, declarations))
            self.error.extend(t.condition.check_type(rty.BOOLEAN))
            self._reference_variable_declaration(t.condition.variables(), declarations)

            t.condition.substituted(lambda e: error_on_unsupported_expression(e, self.error))

        if not state.exception_transition and state.has_exceptions:
            self.error.extend(
                [
                    (
                        f'missing exception transition in state "{state.identifier}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        state.location,
                    ),
                ],
            )

        if state.exception_transition and not state.has_exceptions:
            self.error.extend(
                [
                    (
                        f'unnecessary exception transition in state "{state.identifier}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        state.exception_transition.location,
                    ),
                ],
            )

    def _validate_usage(self) -> None:
        global_declarations = self._global_declarations.items()
        local_declarations = ((k, d) for s in self.states for k, d in s.declarations.items())
        for k, d in itertools.chain(global_declarations, local_declarations):
            if not d.is_referenced:
                self.error.extend(
                    [
                        (
                            f'unused {d.DESCRIPTIVE_NAME} "{k}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            d.location,
                        ),
                    ],
                )

    def _typify_variable(
        self,
        expression: expr.Expr,
        declarations: Mapping[ID, decl.Declaration],
    ) -> expr.Expr:
        if isinstance(
            expression,
            (
                expr.Variable,
                expr.Literal,
                expr.TypeName,
                expr.Call,
                expr.Conversion,
                expr.MessageAggregate,
                expr.DeltaMessageAggregate,
            ),
        ):
            identifier = expression.identifier

            if isinstance(expression, expr.Variable):
                assert identifier not in self._enum_literals
                if identifier in declarations:
                    expression.type_ = declarations[identifier].type_
            if isinstance(expression, expr.Literal) and identifier in self._enum_literals:
                expression.type_ = self._enum_literals[identifier].type_
            if isinstance(expression, expr.TypeName):
                expression.type_ = self._type_names[identifier].type_
            if isinstance(expression, expr.Call) and identifier in declarations:
                expression.type_ = declarations[identifier].type_
                declaration = declarations[identifier]
                assert isinstance(declaration, decl.FunctionDeclaration)
                expression.argument_types = [a.type_ for a in declaration.arguments]
            if isinstance(expression, expr.Conversion) and identifier in self.types:
                expression.type_ = self.types[identifier].type_
                expression.argument_types = [
                    t.pdu.type_
                    for t in self.types.values()
                    if isinstance(t, Refinement) and t.sdu.identifier == identifier
                ]
            if isinstance(expression, expr.MessageAggregate):
                type_identifier = mty.internal_type_identifier(identifier, self.package)
                if type_identifier in self.types:
                    expression.type_ = self.types[type_identifier].type_
            if isinstance(expression, expr.DeltaMessageAggregate) and identifier in declarations:
                expression.type_ = declarations[identifier].type_

        return expression

    @staticmethod
    def _reference_variable_declaration(
        variables: Iterable[expr.Variable],
        declarations: Mapping[ID, decl.Declaration],
    ) -> None:
        for v in variables:
            with contextlib.suppress(KeyError):
                declarations[v.identifier].reference()

    def _validate(self) -> None:
        self._validate_states()

        self._validate_declarations(self._global_declarations, {})

        for s in self.states:
            self._validate_declarations(s.declarations, self._global_declarations)

            declarations = {**self._global_declarations, **s.declarations}

            self._validate_actions(s.actions, declarations, s.declarations)
            self._validate_transitions(s, declarations)

        self._validate_usage()


def normalize_identifiers(
    expression: expr.Expr,
    variables: Iterable[ID],
    enum_literals: Iterable[ID],
    type_names: Iterable[ID],
    functions: Iterable[ID],
) -> expr.Expr:
    variables_map = {v: v for v in variables}
    type_names_map = {t: t for t in type_names}
    enum_literals_map = {l: l for l in enum_literals}
    functions_map = {f: f for f in functions}

    if isinstance(expression, expr.Variable):
        if expression.identifier in type_names_map:
            return expr.TypeName(
                ID(type_names_map[expression.identifier], location=expression.identifier.location),
                expression.type_,
                location=expression.location,
            )
        if expression.identifier in enum_literals_map:
            return expr.Literal(
                ID(
                    enum_literals_map[expression.identifier],
                    location=expression.identifier.location,
                ),
                expression.type_,
                location=expression.location,
            )
        if expression.identifier in functions_map:
            return expr.Call(
                ID(functions_map[expression.identifier], location=expression.identifier.location),
                [],
                expression.immutable,
                expression.type_,
                location=expression.location,
            )
        if expression.identifier in variables_map:
            return expr.Variable(
                ID(variables_map[expression.identifier], location=expression.identifier.location),
                expression.immutable,
                expression.type_,
                location=expression.location,
            )

    if isinstance(expression, expr.Call) and expression.identifier in functions_map:
        return expr.Call(
            ID(functions_map[expression.identifier], location=expression.identifier.location),
            expression.args,
            expression.immutable,
            expression.type_,
            expression.argument_types,
            location=expression.location,
        )

    return expression


@dataclass
class UncheckedSession(UncheckedTopLevelDeclaration):
    identifier: ID
    states: Sequence[State]
    declarations: Sequence[decl.BasicDeclaration]
    parameters: Sequence[decl.FormalDeclaration]
    location: Optional[Location]

    def checked(
        self,
        declarations: Sequence[TopLevelDeclaration],
        skip_verification: bool = False,  # noqa: ARG002
        workers: int = 1,  # noqa: ARG002
    ) -> Session:
        return Session(
            self.identifier,
            deepcopy(self.states),
            deepcopy(self.declarations),
            deepcopy(self.parameters),
            [d for d in declarations if isinstance(d, mty.Type)],
            self.location,
        )


FINAL_STATE: Final[State] = State("Final")


def error_on_unsupported_expression(expression: expr.Expr, error: RecordFluxError) -> expr.Expr:
    # TODO(eng/recordflux/RecordFlux#1497): Support comparisons of opaque fields
    if isinstance(expression, (expr.Equal, expr.NotEqual)):
        for e in [expression.left, expression.right]:
            if isinstance(e, expr.Selected) and e.type_ == rty.OPAQUE:
                error.extend(
                    [
                        (
                            "comparisons of opaque fields not yet supported",
                            Subsystem.MODEL,
                            Severity.ERROR,
                            expression.left.location,
                        ),
                    ],
                )
    return expression



from .bitstring import Bitstring as Bitstring
from .error import PyRFLXError as PyRFLXError
from .package import Package as Package
from .pyrflx import PyRFLX as PyRFLX
from .typevalue import (
    ChecksumFunction as ChecksumFunction,
    EnumValue as EnumValue,
    IntegerValue as IntegerValue,
    MessageValue as MessageValue,
    OpaqueValue as OpaqueValue,
    SequenceValue as SequenceValue,
    TypeValue as TypeValue,
)



from __future__ import annotations

import functools
import inspect
import threading
import uuid
from collections import defaultdict
from pathlib import Path
from typing import Callable, Final, Iterable, Mapping, Optional
from urllib.parse import unquote, urlparse

from lsprotocol.types import (
    TEXT_DOCUMENT_CODE_LENS,
    TEXT_DOCUMENT_DEFINITION,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
    CodeLens,
    CodeLensParams,
    Command,
    DefinitionParams,
    Diagnostic,
    DiagnosticSeverity,
    DidChangeTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    Location,
    Position,
    Range,
    SemanticTokens,
    SemanticTokensLegend,
    SemanticTokensParams,
    TextDocumentItem,
    WorkDoneProgressBegin,
    WorkDoneProgressEnd,
)
from pygls.server import LanguageServer

from rflx import __version__, error
from rflx.common import assert_never
from rflx.const import BUILTINS_PACKAGE, CACHE_PATH, INTERNAL_PACKAGE
from rflx.graph import create_message_graph, write_graph
from rflx.model import Message, Model, UncheckedModel
from rflx.model.cache import Cache
from rflx.specification import Parser

from .lexer import LSLexer
from .model import LSModel

LSP_TOKEN_CATEGORIES: Final = {
    "type": 0,
    "enum": 1,
    "enumMember": 2,
    "struct": 3,
    "property": 4,
    "keyword": 5,
    "class": 6,
    "namespace": 7,
    "event": 8,
    "method": 9,
    "parameter": 10,
    "variable": 11,
}


def to_lsp_location(location: Optional[error.Location]) -> Optional[Location]:
    if (
        location is None
        or location.source is None
        or location.source.name in [str(BUILTINS_PACKAGE), str(INTERNAL_PACKAGE)]
    ):
        return None

    source = location.source.absolute().as_uri()
    start = location.start
    end = location.end if location.end is not None else location.start

    return Location(
        source,
        Range(Position(start[0] - 1, start[1] - 1), Position(end[0] - 1, end[1] - 1)),
    )


def to_lsp_severity(severity: error.Severity) -> Optional[DiagnosticSeverity]:
    if severity is error.Severity.ERROR:
        return DiagnosticSeverity.Error
    if severity is error.Severity.INFO:
        return DiagnosticSeverity.Information
    if severity is error.Severity.WARNING:
        return DiagnosticSeverity.Warning
    if severity is error.Severity.NONE:
        return None
    assert_never(severity)


def initialize_lexer(language_server: RecordFluxLanguageServer, uri: str) -> LSLexer:
    document = language_server.workspace.get_text_document(uri)
    lexer = LSLexer(language_server.models[Path(document.path).parent].ls_model)
    lexer.tokenize(document.source, document.path)
    return lexer


class Models:
    def __init__(self, unchecked_model: UncheckedModel) -> None:
        self.unchecked_model = unchecked_model
        self.checked_model = Model()
        self.ls_model = LSModel(unchecked_model)

    def verify(self, cache: Cache, workers: int) -> None:
        self.checked_model = Model()
        self.checked_model = self.unchecked_model.checked(cache, workers=workers)


class RecordFluxLanguageServer(LanguageServer):
    CMD_SHOW_MESSAGE_GRAPH: Final = "showMessageGraph"

    def __init__(self, workers: int = 1) -> None:
        super().__init__("RecordFlux Language Server", __version__)
        self.workers = workers
        self._cache = Cache()
        self._models: dict[Path, Models] = {}
        self._document_state: dict[str, int] = {}
        self._error: error.RecordFluxError

    @property
    def models(self) -> Mapping[Path, Models]:
        return self._models

    def needs_update_for_document(self, document: TextDocumentItem) -> bool:
        return hash(document.text) != self._document_state.get(document.uri, None)

    def update(self, document_uri: str) -> None:
        token = str(uuid.uuid4())
        self.progress.create(token)
        self.progress.begin(
            token,
            WorkDoneProgressBegin(title="RecordFlux Update", percentage=0, cancellable=False),
        )

        document_path = Path(unquote(urlparse(document_uri).path))
        directory = document_path.parent
        workspace_files = self._workspace_files()
        parser = Parser(self._cache, workers=self.workers)

        self._error = error.RecordFluxError()

        for path in workspace_files:
            document = self.workspace.get_text_document(path.as_uri())
            self._document_state[document.uri] = hash(document.source)
            try:
                parser.parse_string(document.source, path)
            except error.RecordFluxError as e:
                self._error.extend(e)

        unchecked_model = parser.create_unchecked_model()

        self._error.extend(unchecked_model.error)

        self._publish_errors_as_diagnostics(self._error)
        self._reset_diagnostics(
            set(workspace_files) - {e.location.source for e in self._error.errors if e.location},
        )
        self._models[directory] = Models(unchecked_model)

        self.progress.end(token, WorkDoneProgressEnd(message="RecordFlux Update Completed"))

    def verify(self, document_uri: str) -> None:
        token = str(uuid.uuid4())
        self.progress.create(token)
        self.progress.begin(
            token,
            WorkDoneProgressBegin(title="RecordFlux Verification", percentage=0, cancellable=True),
        )

        directory = Path(unquote(urlparse(document_uri).path)).parent

        try:
            self._models[directory].verify(self._cache, workers=self.workers)
        except error.RecordFluxError as e:
            self._error.extend(e)

        self._publish_errors_as_diagnostics(self._error)

        self.progress.end(token, WorkDoneProgressEnd(message="RecordFlux Verification Completed"))

    def _publish_errors_as_diagnostics(self, errors: error.RecordFluxError) -> None:
        diagnostics = defaultdict(list)

        for msg in errors.messages:
            location = to_lsp_location(msg.location)
            severity = to_lsp_severity(msg.severity)

            if location is not None:
                diagnostics[location.uri].append(Diagnostic(location.range, msg.message, severity))

        for uri, diag in diagnostics.items():
            self.publish_diagnostics(uri, diag)

    def _reset_diagnostics(self, files: Iterable[Path]) -> None:
        for f in files:
            self.publish_diagnostics(f.as_uri(), [])

    def _workspace_files(self) -> list[Path]:
        workspace_files = [
            file
            for folder_uri in list(self.workspace.folders)
            for file in Path(unquote(urlparse(folder_uri).path)).rglob("*.rflx")
            if file.is_file()
        ]

        if len(self.workspace.folders) == 0:
            for document_uri in self.workspace.documents:
                document_path = Path(unquote(urlparse(document_uri).path))
                workspace_files.append(document_path)

        return workspace_files


# TODO(eng/recordflux/RecordFlux#1424): Use typing.ParamSpec instead of ... and object
def debounce(  # type: ignore[misc]
    interval_s: int,
    keyed_by: Optional[str] = None,
) -> Callable[[Callable[..., None]], Callable[..., None]]:
    """
    Debounce calls to this function until interval_s seconds have passed.

    Based on decorator from https://github.com/python-lsp/python-lsp-server.
    """

    def wrapper(func: Callable[..., None]) -> Callable[..., None]:  # type: ignore[misc]
        timers: dict[object, threading.Timer] = {}
        lock = threading.Lock()

        @functools.wraps(func)
        def debounced(*args: object, **kwargs: object) -> None:
            sig = inspect.signature(func)
            call_args = sig.bind(*args, **kwargs)
            key = call_args.arguments[keyed_by] if keyed_by else None

            def run() -> None:
                with lock:
                    del timers[key]
                return func(*args, **kwargs)

            with lock:
                old_timer = timers.get(key)
                if old_timer:
                    old_timer.cancel()

                timer = threading.Timer(interval_s, run)
                timers[key] = timer
                timer.start()

        return debounced

    return wrapper


@debounce(1, keyed_by="uri")
def update_model_debounced(ls: RecordFluxLanguageServer, uri: str) -> None:
    ls.update(uri)


@debounce(1, keyed_by="uri")
def update_model_and_verify_debounced(ls: RecordFluxLanguageServer, uri: str) -> None:
    ls.update(uri)
    ls.thread_pool_executor.submit(lambda: ls.verify(uri))


server = RecordFluxLanguageServer()


@server.feature(TEXT_DOCUMENT_DID_OPEN)
async def did_open(ls: RecordFluxLanguageServer, params: DidOpenTextDocumentParams) -> None:
    if ls.needs_update_for_document(params.text_document):
        ls.update(params.text_document.uri)
        ls.thread_pool_executor.submit(lambda: ls.verify(params.text_document.uri))


@server.feature(TEXT_DOCUMENT_DID_SAVE)
async def did_save(ls: RecordFluxLanguageServer, params: DidSaveTextDocumentParams) -> None:
    update_model_and_verify_debounced(ls, params.text_document.uri)


@server.feature(TEXT_DOCUMENT_DID_CHANGE)
async def did_change(ls: RecordFluxLanguageServer, params: DidChangeTextDocumentParams) -> None:
    update_model_debounced(ls, params.text_document.uri)


@server.feature(TEXT_DOCUMENT_DEFINITION)
async def go_to_definition(
    ls: RecordFluxLanguageServer,
    params: DefinitionParams,
) -> Optional[list[Location]]:
    lexer = initialize_lexer(ls, params.text_document.uri)
    token = lexer.search_token(params.position.line, params.position.character)

    if token is None or token.symbol is None:
        return None

    location = to_lsp_location(token.symbol.definition_location)

    if location is None:
        return None

    return [location]


@server.feature(
    TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
    SemanticTokensLegend(token_types=list(LSP_TOKEN_CATEGORIES), token_modifiers=[]),
)
async def semantic_tokens(
    ls: RecordFluxLanguageServer,
    params: SemanticTokensParams,
) -> SemanticTokens:
    lexer = initialize_lexer(ls, params.text_document.uri)

    result: list[int] = []

    previous_line = 0
    previous_offset = 0

    for token in lexer.tokens:
        if token.symbol is None:
            continue

        if token.line_number != previous_line:
            previous_offset = 0

        relative_line = token.line_number - previous_line
        relative_offset = token.character_offset - previous_offset
        length = len(token.lexeme)
        token_category = LSP_TOKEN_CATEGORIES[token.symbol.category.to_lsp_token()]

        previous_line = token.line_number
        previous_offset = token.character_offset

        result.extend([relative_line, relative_offset, length, token_category, 0])

    return SemanticTokens(data=result)


@server.command(RecordFluxLanguageServer.CMD_SHOW_MESSAGE_GRAPH)
async def show_message_graph(ls: RecordFluxLanguageServer, parameters: list[object]) -> None:
    assert isinstance(parameters[0], str)
    assert isinstance(parameters[1], int)

    directory = Path(parameters[0])
    message = ls.models[directory].checked_model.declarations[parameters[1]]

    assert isinstance(message, Message)

    graph_cache = CACHE_PATH / "graphs"
    graph_cache.mkdir(parents=True, exist_ok=True)

    graph = create_message_graph(message)
    write_graph(graph, graph_cache / f"{message.name}.svg")


@server.feature(TEXT_DOCUMENT_CODE_LENS)
async def code_lens(ls: RecordFluxLanguageServer, params: CodeLensParams) -> list[CodeLens]:
    directory = Path(unquote(urlparse(params.text_document.uri).path)).parent

    if directory not in ls.models:
        return []

    result: list[CodeLens] = []

    for index, declaration in enumerate(ls.models[directory].checked_model.declarations):
        if not isinstance(declaration, Message):
            continue

        location = to_lsp_location(declaration.location)

        if location is None or location.uri != params.text_document.uri:
            continue

        result.append(
            CodeLens(
                location.range,
                Command(
                    "Show message graph",
                    RecordFluxLanguageServer.CMD_SHOW_MESSAGE_GRAPH,
                    [directory, index],
                ),
            ),
        )

    return result



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Generator, Mapping, Sequence
from typing import Optional

from rflx import ir, typing_ as rty
from rflx.common import Base
from rflx.error import Location, RecordFluxError, Severity, Subsystem
from rflx.expression import Expr, Variable, _to_ir_basic_expr
from rflx.identifier import ID, StrID


class Statement(Base):
    def __init__(
        self,
        identifier: StrID,
        type_: rty.Type = rty.UNDEFINED,
        location: Optional[Location] = None,
    ):
        self.identifier = ID(identifier)
        self.type_ = type_
        self.location = location

    @abstractmethod
    def check_type(
        self,
        statement_type: rty.Type,
        typify_variable: Callable[[Expr], Expr],
    ) -> RecordFluxError:
        """Set the types of variables, and check the types of the statement and expressions."""
        raise NotImplementedError

    @abstractmethod
    def variables(self) -> Sequence[Variable]:
        """Return all referenced variables in the statement."""
        raise NotImplementedError

    @abstractmethod
    def to_ir(self, variable_id: Generator[ID, None, None]) -> list[ir.Stmt]:
        raise NotImplementedError


class Assignment(Statement):
    def __init__(
        self,
        identifier: StrID,
        expression: Expr,
        type_: rty.Type = rty.UNDEFINED,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, type_, location)
        self.expression = expression

    def variables(self) -> Sequence[Variable]:
        return [Variable(self.identifier), *self.expression.variables()]


class VariableAssignment(Assignment):
    def __str__(self) -> str:
        return f"{self.identifier} := {self.expression}"

    def check_type(
        self,
        statement_type: rty.Type,
        typify_variable: Callable[[Expr], Expr],
    ) -> RecordFluxError:
        self.type_ = statement_type
        self.expression = self.expression.substituted(typify_variable)
        return rty.check_type_instance(
            statement_type,
            rty.Any,
            self.location,
            f'variable "{self.identifier}"',
        ) + self.expression.check_type(statement_type)

    def to_ir(self, variable_id: Generator[ID, None, None]) -> list[ir.Stmt]:
        assert isinstance(self.type_, rty.NamedType)
        expression = self.expression.to_ir(variable_id)
        return [*expression.stmts, ir.Assign(self.identifier, expression.expr, self.type_, self)]


class MessageFieldAssignment(Assignment):
    def __init__(
        self,
        message: StrID,
        field: StrID,
        expression: Expr,
        type_: rty.Type = rty.UNDEFINED,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(message, expression, type_, location)
        self.message = ID(message)
        self.field = ID(field)

    def __str__(self) -> str:
        return f"{self.message}.{self.field} := {self.expression}"

    def check_type(
        self,
        statement_type: rty.Type,
        typify_variable: Callable[[Expr], Expr],
    ) -> RecordFluxError:
        error = RecordFluxError()
        field_type: rty.Type = rty.UNDEFINED
        if isinstance(statement_type, rty.Message):
            if self.field in statement_type.fields:
                field_type = statement_type.types[self.field]
            elif self.field in statement_type.parameters:
                error.extend(
                    [
                        (
                            f'message parameter "{self.field}" cannot be set using an assignment',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            self.field.location,
                        ),
                        (
                            "use a Reset statement to change the message parameters",
                            Subsystem.MODEL,
                            Severity.INFO,
                            self.field.location,
                        ),
                    ],
                )
            else:
                error.extend(
                    [
                        (
                            f'invalid message field "{self.field}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            self.field.location,
                        ),
                    ],
                )
        self.type_ = statement_type
        self.expression = self.expression.substituted(typify_variable)
        return (
            error
            + rty.check_type_instance(
                statement_type,
                rty.Message,
                self.location,
                f'variable "{self.identifier}"',
            )
            + self.expression.check_type(field_type)
        )

    def to_ir(self, variable_id: Generator[ID, None, None]) -> list[ir.Stmt]:
        assert isinstance(self.type_, rty.Message)

        expression = self.expression.to_ir(variable_id)
        return [
            *expression.stmts,
            ir.FieldAssign(self.message, self.field, expression.expr, self.type_, self),
        ]


class AttributeStatement(Statement):
    def __init__(
        self,
        identifier: StrID,
        attribute: str,
        parameters: list[Expr],
        type_: rty.Type = rty.UNDEFINED,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, type_, location)
        self.attribute = attribute
        self.parameters = parameters

    def __str__(self) -> str:
        parameters = ", ".join([str(p) for p in self.parameters])
        return f"{self.identifier}'{self.attribute}" + (f" ({parameters})" if parameters else "")

    def check_type(
        self,
        statement_type: rty.Type,
        typify_variable: Callable[[Expr], Expr],
    ) -> RecordFluxError:
        raise NotImplementedError

    def variables(self) -> Sequence[Variable]:
        return [Variable(self.identifier), *[e for p in self.parameters for e in p.variables()]]

    def to_ir(self, variable_id: Generator[ID, None, None]) -> list[ir.Stmt]:
        raise NotImplementedError


class ListAttributeStatement(AttributeStatement):
    def __init__(
        self,
        identifier: StrID,
        parameter: Expr,
        type_: rty.Type = rty.UNDEFINED,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, self.__class__.__name__, [parameter], type_, location)


class Append(ListAttributeStatement):
    def check_type(
        self,
        statement_type: rty.Type,
        typify_variable: Callable[[Expr], Expr],
    ) -> RecordFluxError:
        self.type_ = statement_type
        self.parameter = self.parameter.substituted(typify_variable)
        error = rty.check_type_instance(
            statement_type,
            rty.Sequence,
            self.location,
            f'variable "{self.identifier}"',
        )
        if isinstance(statement_type, rty.Sequence):
            error += self.parameter.check_type(statement_type.element)
            if isinstance(statement_type.element, rty.Message) and isinstance(
                self.parameter,
                Variable,
            ):
                error.extend(
                    [
                        (
                            "appending independently created message not supported",
                            Subsystem.MODEL,
                            Severity.ERROR,
                            self.parameter.location,
                        ),
                        (
                            "message aggregate should be used instead",
                            Subsystem.MODEL,
                            Severity.INFO,
                            self.parameter.location,
                        ),
                    ],
                )
        return error

    @property
    def parameter(self) -> Expr:
        return self.parameters[0]

    @parameter.setter
    def parameter(self, value: Expr) -> None:
        assert isinstance(self.parameters, list)
        self.parameters[0] = value

    def to_ir(self, variable_id: Generator[ID, None, None]) -> list[ir.Stmt]:
        assert isinstance(self.type_, rty.Sequence)
        parameter = self.parameter.to_ir(variable_id)
        return [
            *parameter.stmts,
            ir.Append(self.identifier, parameter.expr, self.type_, self),
        ]


class Extend(ListAttributeStatement):
    def check_type(
        self,
        statement_type: rty.Type,
        typify_variable: Callable[[Expr], Expr],
    ) -> RecordFluxError:
        self.type_ = statement_type
        self.parameter = self.parameter.substituted(typify_variable)
        return rty.check_type_instance(
            statement_type,
            rty.Sequence,
            self.location,
            f'variable "{self.identifier}"',
        ) + self.parameter.check_type(statement_type)

    @property
    def parameter(self) -> Expr:
        return self.parameters[0]

    @parameter.setter
    def parameter(self, value: Expr) -> None:
        assert isinstance(self.parameters, list)
        self.parameters[0] = value

    def to_ir(self, variable_id: Generator[ID, None, None]) -> list[ir.Stmt]:
        assert isinstance(self.type_, rty.Sequence)
        parameter = self.parameter.to_ir(variable_id)
        return [
            *parameter.stmts,
            ir.Extend(self.identifier, parameter.expr, self.type_, self),
        ]


class Reset(AttributeStatement):
    def __init__(
        self,
        identifier: StrID,
        associations: Optional[Mapping[ID, Expr]] = None,
        type_: rty.Type = rty.UNDEFINED,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, self.__class__.__name__, [], type_, location)
        self.associations = associations or {}

    def __str__(self) -> str:
        associations = ", ".join([f"{k} => {v}" for k, v in self.associations.items()])
        return f"{self.identifier}'{self.attribute}" + (
            f" ({associations})" if associations else ""
        )

    def check_type(
        self,
        statement_type: rty.Type,
        typify_variable: Callable[[Expr], Expr],
    ) -> RecordFluxError:
        error = RecordFluxError()
        self.type_ = statement_type
        self.associations = {
            i: e.substituted(typify_variable) for i, e in self.associations.items()
        }
        if isinstance(statement_type, rty.Sequence):
            for i, e in self.associations.items():
                error.extend(
                    [
                        (
                            f'unexpected argument "{i}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            e.location,
                        ),
                    ],
                )
        elif isinstance(statement_type, rty.Message):
            for i, e in self.associations.items():
                if i in statement_type.parameter_types:
                    error.extend(e.check_type(statement_type.parameter_types[i]))
                else:
                    error.extend(
                        [
                            (
                                f'unexpected argument "{i}"',
                                Subsystem.MODEL,
                                Severity.ERROR,
                                e.location,
                            ),
                        ],
                    )
            for a in statement_type.parameter_types:
                if a not in self.associations:
                    error.extend(
                        [
                            (
                                f'missing argument "{a}"',
                                Subsystem.MODEL,
                                Severity.ERROR,
                                self.location,
                            ),
                        ],
                    )
        return error + rty.check_type_instance(
            statement_type,
            (rty.Sequence, rty.Message),
            self.location,
            f'variable "{self.identifier}"',
        )

    def variables(self) -> Sequence[Variable]:
        return [
            Variable(self.identifier),
            *[v for e in self.associations.values() for v in e.variables()],
        ]

    def to_ir(self, variable_id: Generator[ID, None, None]) -> list[ir.Stmt]:
        assert isinstance(self.type_, (rty.Sequence, rty.Message))
        associations = {}
        stmts = []
        for i, e in self.associations.items():
            e_ir = e.to_ir(variable_id)
            associations[i] = e_ir.expr
            stmts.extend(e_ir.stmts)
        return [
            *stmts,
            ir.Reset(self.identifier, associations, self.type_, self),
        ]


class ChannelAttributeStatement(AttributeStatement):
    def __init__(
        self,
        identifier: StrID,
        parameter: Expr,
        type_: rty.Type = rty.UNDEFINED,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, self.__class__.__name__, [parameter], type_, location)

    def check_type(
        self,
        statement_type: rty.Type,
        typify_variable: Callable[[Expr], Expr],
    ) -> RecordFluxError:
        self.type_ = statement_type
        self.parameters = [self.parameter.substituted(typify_variable)]
        return rty.check_type(
            statement_type,
            self._expected_channel_type(),
            self.location,
            f'channel "{self.identifier}"',
        ) + self.parameter.check_type_instance(rty.Message)

    @property
    def parameter(self) -> Expr:
        return self.parameters[0]

    def to_ir(self, variable_id: Generator[ID, None, None]) -> list[ir.Stmt]:
        parameter_stmts, parameter_expr = _to_ir_basic_expr(self.parameter, variable_id)
        return [
            *parameter_stmts,
            getattr(ir, self.__class__.__name__)(self.identifier, parameter_expr, self),
        ]

    @abstractmethod
    def _expected_channel_type(self) -> rty.Channel:
        raise NotImplementedError


class Read(ChannelAttributeStatement):
    def _expected_channel_type(self) -> rty.Channel:
        return rty.Channel(readable=True, writable=False)


class Write(ChannelAttributeStatement):
    def _expected_channel_type(self) -> rty.Channel:
        return rty.Channel(readable=False, writable=True)
